from chitra.utility.import_utils import is_installed

if is_installed("loguru"):
    from loguru import logger

    logger.debug("Using loguru for logging!")
else:
    import logging

    logger = logging.getLogger()



import numpy as np
import tensorflow as tf

from chitra.dataloader import Clf

shape = (32, 32)
class_names = ("a", "b")
clf = Clf()
clf.shape = shape
clf.CLASS_NAMES = class_names


def test__encode_classes():
    clf._encode_classes()
    assert clf.class_to_idx == {"a": 0, "b": 1}


def test__ensure_shape():
    img, label = tf.random.normal((32, 32, 3)), 0
    img1, label1 = clf._ensure_shape(img, label)
    assert np.all(img1 == img)
    assert np.all(label1 == label)


def test__get_image_list():
    assert len(clf._get_image_list(".")) != 0



from tensorflow_serving.apis.prediction_service_pb2_grpc import PredictionServiceStub

from chitra.serve.tf_serving.client import GrpcClient, create_grpc_stub


def test_create_grpc_stub():
    assert isinstance(create_grpc_stub(), PredictionServiceStub)


def test_request():
    client = GrpcClient()
    assert isinstance(client.stub, PredictionServiceStub)



import numpy as np
from fastapi import FastAPI

from chitra.serve import create_api
from chitra.serve.model_server import _default_postprocess


def test_create_app():
    model = lambda x: x + 1
    app = create_api(model)
    assert isinstance(app, FastAPI)


def test__default_postprocess():
    data = _default_postprocess(np.random.randn(5))
    assert len(data) == 5
    assert isinstance(data, list)



from typing import List, Optional, Tuple, Union

import numpy as np
from imgaug.augmentables import bbs


class BoundingBoxes:
    CENTER = "XXYY"
    CORNER = "XYXY"

    def __init__(
        self,
        bboxes: Optional[List[list]] = None,
        labels: Optional[List[Union[int, str]]] = None,
        box_format: str = "xyxy",
    ):
        """Args:
        bboxes: list of bounding boxes [(x1, y1, x2, y2), ...] or [(xc, yc, h, w), ...]
        labels: list of strings or integers
        box_format:
            - `xyxy` for corner points of bbox
            - `xyhw` for x-center, y-center, height and width format of bbox
        """
        if box_format.upper() not in (
            self.CENTER,
            self.CORNER,
        ):
            raise AssertionError("bbox format must be either xyxy or xyhw")
        bboxes = self._listify(bboxes, 4)
        labels = self._listify(labels)

        if len(bboxes) != len(labels):
            raise UserWarning(
                f"len of boxes and labels not matching: {len(bboxes), len(labels)}"
            )
        self._format = box_format.upper()
        self.bboxes = self._list_to_bbox(bboxes, labels)
        self._state = {}

    def _listify(self, item, dim_trigger=None):
        if item is None:
            return item

        if not isinstance(item, (list, tuple)):
            return [item]

        if isinstance(item, (list, tuple)) and self.num_dim(item) == dim_trigger:
            item = [item]
        return item

    @staticmethod
    def num_dim(item):
        return len(item)

    @staticmethod
    def center_to_corner(cx, cy, h, w):
        xmin = cx - w / 2
        xmax = cx + w / 2
        ymin = cy - h / 2
        ymax = cy + h / 2

        return xmin, ymin, xmax, ymax

    @staticmethod
    def corner_to_center(xmin, ymin, xmax, ymax):
        w = xmax - xmin
        h = ymax - ymin

        cx = xmin + w / 2
        cy = ymin + h / 2

        return cx, cy, h, w

    def _list_to_bbox(
        self,
        bbox_list: Optional[List[List[Union[int, float]]]],
        labels: List[Union[str, int]] = None,
    ) -> List[bbs.BoundingBox]:
        """Converts bbox list into `imgaug BoundigBox` object"""
        _format = self._format

        if not bbox_list:
            return []

        if not labels:
            labels = [None] * self.num_dim(bbox_list)

        bbox_objects = []
        for bbox, label in zip(bbox_list, labels):
            if _format == self.CENTER:
                bbox = self.center_to_corner(*bbox)
            bbox_objects.append(bbs.BoundingBox(*bbox, label))
        return bbox_objects

    def __getitem__(self, idx):
        return self.bboxes[idx]

    def __repr__(self):
        return str(self.bboxes)

    def get_bounding_boxes_on_image(
        self, image_shape: Tuple[int]
    ) -> bbs.BoundingBoxesOnImage:
        """returns `imgaug BoundingBoxesOnImage` object which can be used to boxes on the image"""
        return bbs.BoundingBoxesOnImage(self.bboxes, image_shape)

    def resize_with_image(
        self, old_image_size: List[int], rescaled_image: np.ndarray
    ) -> bbs.BoundingBoxesOnImage:
        bbox_on_image = self.get_bounding_boxes_on_image(old_image_size)
        bbox_on_rescaled_image = bbox_on_image.on(rescaled_image)
        self.bboxes = bbox_on_rescaled_image.bounding_boxes
        return bbox_on_rescaled_image



import rich
import numpy as np

from chitra.coordinates import BoundingBoxes

box = [1, 2, 3, 4]
label = ["Dog"]
bounding_box = BoundingBoxes(box, label)
rich.print(bounding_box)
bboxes = bounding_box.resize_with_image((10, 10, 3), np.random.randn(100, 100, 3))

rich.print(bboxes)



import os

import tensorflow as tf


def disable_gpu():
    """
    disable gpu for tensorflow
    """
    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
    os.environ["CUDA_VISIBLE_DEVICES"] = str(-1)


def limit_gpu(gpu_id: int, memory_limit: int):
    """
    limit the selected gpu [gpu_id] by [memory_limit] MB
    """
    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
    os.environ["CUDA_VISIBLE_DEVICES"] = str(gpu_id)
    gpus = tf.config.list_physical_devices("GPU")

    if len(gpus) >= gpu_id + 1:
        raise AssertionError
    if gpus:
        # Restrict TensorFlow to only allocate [memory MB] of memory on the first GPU
        try:
            tf.config.experimental.set_virtual_device_configuration(
                gpus[gpu_id],
                [
                    tf.config.experimental.VirtualDeviceConfiguration(
                        memory_limit=memory_limit
                    )
                ],
            )
            logical_gpus = tf.config.list_logical_devices("GPU")
            print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")
        except RuntimeError as e:
            # Virtual devices must be set before GPUs have been initialized
            print(e)
    else:
        print(f"No GPU:{gpu_id} found in your system!")


def gpu_dynamic_mem_growth():
    """
    Borrowed from https://github.com/philipperemy/keract/tree/master/examples

    Check for GPUs and set them to dynamically grow memory as needed
    Avoids OOM from tensorflow greedily allocating GPU memory
    """
    try:
        gpu_devices = tf.config.list_physical_devices("GPU")
        if len(gpu_devices) > 0:
            for gpu in gpu_devices:
                tf.config.experimental.set_memory_growth(gpu, True)
            print("GPU dynamic memory growth enabled")
        else:
            print("No GPU found on the machine!")
    except AttributeError:
        print(
            "Upgrade your tensorflow to 2.x to have the gpu_dynamic_mem_growth feature."
        )



import os

IMAGENET_LABEL_URL = (
    "https://storage.googleapis.com/download.tensorflow.org/data/ImageNetLabels.txt"
)

_TF = "tensorflow"
_TF_GPU = "tensorflow-gpu"
_TORCH = "torch"
_TORCHVISION = "torchvision"
_JAX = "jax"
_FLAX = "flax"

CHITRA_URL_SEP = "[-]"
CACHE_DIR = os.environ.get("HOME", ".") + "/.chitra_cache"
IMAGE_CACHE_DIR = f"{CACHE_DIR}/images/"



from typing import Union

import tensorflow as tf


def read_image(path: str, channels: int = 3):
    """Reads an image file from the path and return the rgb image in tf.Tensor format."""
    img: tf.Tensor = tf.io.read_file(path)
    img: tf.Tensor = tf.io.decode_image(img, channels=channels, expand_animations=False)
    return img


def resize_image(image: tf.Tensor, size: Union[tf.Tensor, tuple], **kwargs):
    """Resize image to the target `size`: Union[tf.Tensor, tuple]"""
    if not isinstance(image, tf.Tensor):
        raise AssertionError(
            f"image must be of type tf.Tensor but passed {type(image)}"
        )
    if not isinstance(size, (tuple, tf.Tensor)):
        raise AssertionError
    method = kwargs.get("method", "bilinear")
    return tf.image.resize(image, size, method)



import importlib

from chitra.constants import _FLAX, _JAX, _TF, _TF_GPU, _TORCH, _TORCHVISION


def is_installed(module_name: str):
    return importlib.util.find_spec(module_name) is not None


INSTALLED_MODULES = {
    module: is_installed(module)
    for module in (_TF, _TF_GPU, _TORCH, _TORCHVISION, _JAX, _FLAX)
}



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['chitra',
 'chitra.converter',
 'chitra.serve',
 'chitra.serve.schema',
 'chitra.serve.tf_serving',
 'chitra.serve.torch_serve',
 'chitra.utility',
 'chitra.visualization']

package_data = \
{'': ['*']}

install_requires = \
['tensorflow >= 2.3',
 'tensorflow-addons>=0.13.0',
 'tf-keras-vis>=0.5.3',
 'matplotlib',
 'pillow',
 'imgaug >=0.4.0',
 'requests >=2.24.0,<3.0.0',
 'scikit-learn']

extras_require = \
{'all': ['onnx',
         'onnx2pytorch',
         'tf2onnx',
         'fastapi',
         'uvicorn',
         'pydantic',
         'tensorflow-serving-api',
         'torch']}

setup(name='chitra',
      version='0.1.0b3',
      description='A Deep Learning Computer Vision Utility library',
      author='Aniket Maurya',
      author_email='hello@aniketmaury.com',
      url='https://github.com/aniketmaurya/chitra',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.7',
     )



from unittest.mock import MagicMock, Mock, patch

import numpy as np
import pytest

from chitra.visualization.metrics import (
    cm_accuracy,
    detect_multilabel,
    plot_confusion_matrix,
)


def test_detect_multilabel():
    with pytest.raises(UserWarning):
        detect_multilabel({"label1": "this will raise UserWarning"})

    assert detect_multilabel([1, 2, 3, 4])
    assert not detect_multilabel([0, 1, 1, 0])


def test_cm_accuracy():
    x = np.asarray([[1, 2], [1, 2]])
    assert cm_accuracy(x) == 0.5


@patch("chitra.visualization.metrics.plt")
def test_plot_confusion_matrix(mock_plt: Mock):
    mock_plt.show = MagicMock()

    y_pred = [1, 1, 0, 1]
    y_true = [0, 1, 0, 1]

    assert plot_confusion_matrix(y_pred, y_true) is None
    mock_plt.show.assert_called_once()



from functools import partial
from io import BytesIO
from typing import Callable, List, Optional, Tuple

import numpy as np
import uvicorn
from fastapi import FastAPI, File, UploadFile
from PIL import Image

from chitra.serve.schema import QnARequest, QnAResponse, Query, QueryResult

IMAGE_CLF = "IMAGE-CLASSIFICATION"
TXT_CLF = "TEXT-CLASSIFICATION"
QNA = "QUESTION-ANS"
API_TYPES = (IMAGE_CLF, TXT_CLF, QNA)


def _default_preprocess(
    image_file, image_size: Tuple[int, int], rescale: bool
) -> np.ndarray:
    image = Image.open(BytesIO(image_file)).convert("RGB")
    image = image.resize(image_size)
    image = np.asarray(image).astype(np.float32)
    if rescale:
        image = image / 127.5 - 1.0
    image = np.expand_dims(image, 0)
    return image


def _default_postprocess(data) -> List:
    if not isinstance(data, (np.ndarray, list, tuple)):
        data = data.numpy()
    data = data.tolist()
    return data


def create_api(
    model: Callable,
    api_type: str = "IMAGE-CLASSIFICATION",
    preprocess: Optional[Callable] = None,
    postprocess: Optional[Callable] = None,
    run: bool = False,
    **kwargs
) -> FastAPI:
    """
    Create ASGI API with using FastAPI.
    Args:
        model[Callable]: any callable model -> pred = model(x)
        api_type[str]: Type of API to be created -> {"IMAGE-CLASSIFICATION", "TEXT-CLASSIFICATION", "QUESTION-ANS"}
        preprocess[Callable]: Preprocessing function for your data. (Default fn will be applied if None.)
        postprocess[Callable]: Postprocessing function for your data. (Default fn will be applied if None.)
        run[bool]: Whether to the the API or not.
        **kwargs:
            image_size[tuple]: if the api type is image-classification then size of target image.

    Returns:
        FastAPI app

    """
    api_type = api_type.upper()
    docs_url = kwargs.get("docs_url", "/docs")
    title = kwargs.get("title", "Chitra Model Server 🔥")
    desc = kwargs.get(
        "description",
        '<a href="https://chitra.readthedocs.io/en/latest">Goto Chitra Docs</a> 🔗',
    )
    app = FastAPI(title=title, description=desc, docs_url=docs_url)

    @app.get("/healthz")
    def health():
        return {"OK": True}

    if api_type == IMAGE_CLF:

        @app.post("/api/predict-image")
        async def predict_image(file: UploadFile = File(...)):
            preprocess_fn = preprocess
            postprocess_fn = postprocess
            if preprocess is None:
                preprocess_fn = partial(
                    _default_preprocess,
                    image_size=kwargs.get("image_size", (224, 224)),
                    rescale=kwargs.get("rescale", True),
                )
            if postprocess_fn is None:
                postprocess_fn = _default_postprocess

            x = preprocess_fn(await file.read())
            x = model(x)
            x = postprocess_fn(x)
            return x

        if api_type == TXT_CLF:

            @app.post("/api/predict-text", response_model=QueryResult)
            async def predict_text(data: Query):
                x = data.query
                if preprocess:
                    x = preprocess(x)
                x = model(x)
                if postprocess:
                    x = postprocess(x)
                return x

    if api_type == QNA:

        @app.post("/api/QnA", response_model=QnAResponse)
        async def predict_question_answer(data: QnARequest):
            query = data.query
            question = data.question
            if preprocess:
                x = preprocess(query, question)
            x = model(x)
            if postprocess:
                x = postprocess(x)
            return x

    if run:
        uvicorn.run(app)
    return app



import os
import pathlib
import random
import time
from functools import partial
from glob import glob
from pathlib import Path
from typing import Callable, Union

import tensorflow as tf
from typeguard import check_argument_types, typechecked

from .tf_image import read_image, resize_image


def benchmark(dataset, num_epochs=2, fake_infer_time=0.001):
    """Use this function to benchmark your Dataset loading time"""
    start_time = time.perf_counter()
    for _ in range(num_epochs):
        for _ in dataset:
            # Performing a training step
            time.sleep(fake_infer_time)
    tf.print(
        f"Execution time for {num_epochs} epochs: {time.perf_counter() - start_time :0.3f} seconds"
    )


def get_filenames(root_dir):
    root_dir = pathlib.Path(root_dir)
    return glob(str(root_dir / "*/*"))


def get_label(filename):
    return filename.split("/")[-2]


class ImageSizeList:
    def __init__(self, img_sz_list=None):

        if (
            isinstance(img_sz_list, (list, tuple))
            and len(img_sz_list) != 0
            and not isinstance(img_sz_list[0], (list, tuple))
        ):
            img_sz_list = [img_sz_list][:]

        self.start_size = None
        self.last_size = None
        self.curr_size = None
        self.img_sz_list = img_sz_list

        try:
            self.start_size = img_sz_list[0]
            self.last_size = img_sz_list[-1]
            self.curr_size = img_sz_list[0]
        except (IndexError, TypeError):
            print("No item present in the image size list")
            self.curr_size = None  # no item present in the list

    def get_size(self):
        img_sz_list = self.img_sz_list
        try:
            self.curr_size = img_sz_list.pop(0)
        except (IndexError, AttributeError):
            print(f"Returning the last set size which is: {self.curr_size}")

        return self.curr_size


# Cell
class Pipeline:
    @typechecked
    def __init__(self, funcs: Union[Callable, list, tuple] = None):
        if not check_argument_types():
            raise AssertionError
        if isinstance(funcs, list):
            self.funcs = funcs
        elif callable(funcs):
            self.funcs = [funcs]
        else:
            self.funcs = []

    @typechecked
    def add(self, func: Callable):
        if not check_argument_types():
            raise AssertionError
        self.funcs.append(func)

    def __call__(self, item):
        try:
            for func in self.funcs:
                item = func(item)
        except Exception as e:
            print("Error while applying function in pipeline!")
            raise e
        return item


class Dataset:
    MAPPINGS = {
        "PY_TO_TF": {str: tf.string, int: tf.int32, float: tf.float32},
    }

    def __init__(
        self,
        train_dir: Union[str, Path],
        image_size=None,
        transforms=None,
        default_encode=True,
        **kwargs,
    ):
        """
        Create a Dataset object that can generate tf.data.Dataset
        Args:
            train_dir:
            image_size:
            transforms:
            default_encode:
            **kwargs:
        """
        self.get_filenames = get_filenames
        self.read_image = read_image
        self.get_label = get_label
        self.transforms = transforms

        self.root_dir = train_dir
        self.default_encode = default_encode
        self.filenames = self.get_filenames(train_dir)
        self.num_files = len(self.filenames)
        self.image_size = image_size
        self.img_sz_list = ImageSizeList(self.image_size)

        self.labels = kwargs.get("labels", self.get_labels())

    def __len__(self):
        return len(self.filenames)

    def _process(self, filename):
        image = self.read_image(filename)
        label = self.get_label(filename)
        return image, label

    def _reload(self):
        image_size = self.image_size[:]
        self.filenames = self.get_filenames(self.root_dir)
        self.num_files = len(self.filenames)
        self.img_sz_list = ImageSizeList(image_size)
        self.labels = self.get_labels()

    def _capture_return_types(self):
        return_types = []
        for e in self.generator():
            outputs = e
            break
        if isinstance(outputs, tuple):
            for ret_type in outputs:
                return_types.append(
                    ret_type.dtype
                    if tf.is_tensor(ret_type)
                    else Dataset.MAPPINGS["PY_TO_TF"][type(ret_type)]
                )
        else:
            raise UserWarning("Unable to capture return type!")
        return tuple(return_types)

    def __getitem__(self, idx):
        filename = self.filenames[idx]
        return self._process(filename)

    def update_component(self, component_name, new_component, reload=True):
        setattr(self, component_name, new_component)
        print(f"{component_name} updated with {new_component}")
        self._reload()

    def get_labels(self):
        # get labels should also update self.num_classes
        root_dir = self.root_dir
        labels = set()
        folders = glob(f"{root_dir}/*")
        for folder in folders:
            labels.add(os.path.basename(folder))

        labels = sorted(labels)
        self.NUM_CLASSES = len(labels)
        self.label_to_idx = {label: i for i, label in enumerate(labels)}

        return labels

    def label_encoder(self, label):
        idx = self.label_to_idx.get(label, None)
        if idx is None:
            raise AssertionError(f"Error while converting label={label} to index!")
        return idx

    def generator(self, shuffle=False):
        if shuffle:
            random.shuffle(self.filenames)
        img_sz = self.img_sz_list.get_size()
        n = len(self.filenames)
        for i in range(n):
            image, label = self.__getitem__(i)
            if img_sz:
                image = resize_image(image, img_sz)
            if self.transforms:
                image = self.transforms(image)
            if self.default_encode is True:
                label = self.label_encoder(label)
            yield image, label

    def get_tf_dataset(self, output_shape=None, shuffle=True):
        return_types = self._capture_return_types()
        self._reload()
        generator = partial(self.generator, shuffle=shuffle)
        datagen = tf.data.Dataset.from_generator(generator, return_types, output_shape)

        return datagen
